#
# Copyright (c) 1995 The Regents of the University of California.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#	This product includes software developed by the Computer Systems
#	Engineering Group at Lawrence Berkeley Laboratory.
# 4. Neither the name of the University nor of the Laboratory may be used
#    to endorse or promote products derived from this software without
#    specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# This test produces output corresponding to one-way traffic over a 
# geosynchronous satellite.  Three transport protocols can be used:
# STP (stp), TCP-Sack (sack), and TCP-Reno (newreno).
# This tcl script is based on the ns "test_suite.tcl" test script.
#
# To run individual tests:
# ns test1g.tcl <stp, sack, or newreno>
# ...
#

source httpModel.tcl

Class TestSuite

TestSuite instproc init {} {
	$self instvar ns_ net_ defNet_ test_ topo_ node_ testName_
	set ns_ [new Simulator]
	# trace-all is only used in more extensive test suites
	# $ns_ trace-all [open all.tr w]
	if {$net_ == ""} {
		set net_ $defNet_
	}
	if ![Topology/$defNet_ info subclass Topology/$net_] {
		global argv0
		puts "$argv0: cannot run test $test_ over topology $net_"
		exit 1
	}

	set topo_ [new Topology/$net_ $ns_]
	foreach i [$topo_ array names node_] {
		# This would be cool, but lets try to be compatible
		# with test-suite.tcl as far as possible.
		#
		# $self instvar $i
		# set $i [$topo_ node? $i]
		#
		set node_($i) [$topo_ node? $i]
	}

	if {$net_ == $defNet_} {
		set testName_ "$test_"
	} else {
		set testName_ "$test_:$net_"
	}

	# XXX
	if [info exists node_(k1)] {
		set blink [$ns_ link $node_(r1) $node_(k1)]
	} else {
		set blink [$ns_ link $node_(r1) $node_(r2)] 
	}
	$blink trace-dynamics $ns_ stdout
}

TestSuite instproc finish file {
	$self instvar tcp1
	global env

	set data_sent [$tcp1 set ack_]
   
	#
	# we don't bother checking for the link we're interested in
	# since we know only such events are in our trace file
	#
exec awk {
        {
            if ($1 == "+" && $5 == "tcp" && ($9 == "0.0" || $9 == "0.1") && ($10 == "5.0" || $10 == "5.1"))
                print $2, $6  
        }
    } out.tr > out.tr1
    
exec awk {
		{
            if ($1 == "+" && $5 == "ack" && ($9 == "5.0" || $9 == "5.1") && ($10 == "0.0" || $10 == "0.1"))
                print $2, $6 
        }
    } out.tr > out.tr2

set forward [exec awk { {s = s + $2} END {print s} } out.tr1]  
set reverse [exec awk { {s = s + $2} END {print s} } out.tr2]
set forward_data [expr $data_sent * 960.0]
set throughput [expr $forward_data * 8 / 59.4]
set efficiency [expr $forward_data / $forward]
set reverse_channel [expr $reverse * 8 / 60.0]
puts "$throughput $efficiency $reverse_channel"
    
	if [info exists env(DISPLAY)] {
#	    exec Xgraph -display $env(DISPLAY) -bb -tk -nl -m -x time -y packet temp.rands &
	} else {
	    puts stderr "output trace is in temp.rands"
	}
	
#	exit 0
}

TestSuite instproc finish_stp file {
	$self instvar stp1
	global env

	set data_sent [$stp1 set highest_ack_]
	set fname [ns-random 0]
    set fwddatatmp $fname.fwddata ; # data-full acks
    set pollstmp $fname.polls; # pure acks
    set statstmp $fname.stats
   
	#
	# we don't bother checking for the link we're interested in
	# since we know only such events are in our trace file
	#
exec awk {
        {
            if ($1 == "+" && ($5 == "sd" || $5 == "poll" )&& ($9 == "0.0" || $9 == "0.1") && ($10 == "5.0" || $10 == "5.1"))
                print $2, $6  
        }
    } out.tr > out.tr1
    
exec awk {
		{
            if ($1 == "+" && ($5 == "stat" || $5 == "ustat") && ($9 == "5.0" || $9 == "5.1") && ($10 == "0.0" || $10 == "0.1"))
                print $2, $6 
        }
    } out.tr > out.tr2

set forward [exec awk { {s = s + $2} END {print s} } out.tr1]  
set reverse [exec awk { {s = s + $2} END {print s} } out.tr2]
set forward_data [expr $data_sent * 968.0]
# Because we calculate the forward throughput based on the highest ack number
# received, we must account for the approx. 0.6 seconds of round trip latency
# in this number (this accounts for dividing the forward data by 59.4 
# instead of 60.0).  This is a quick hack and must be consistent with the
# TCP calculations as well.
set throughput [expr $forward_data * 8 / 59.4]
set efficiency [expr $forward_data / $forward]
set reverse_channel [expr $reverse * 8 / 60.0]
puts "$throughput $efficiency $reverse_channel"
    
	if [info exists env(DISPLAY)] {
#	    exec Xgraph -display $env(DISPLAY) -bb -tk -nl -m -x time -y packet temp.rands &
	} else {
	    puts stderr "output trace is in temp.rands"
	}
	
#	exit 0
}

#
# Arrange for stp source stats to be dumped for $tcpSrc every
# $interval seconds of simulation time
#
TestSuite instproc stpDump { stpSrc interval } {
	$self instvar dump_inst_ ns_
	if ![info exists dump_inst_($stpSrc)] {
		set dump_inst_($stpSrc) 1
		$ns_ at 0.0 "$self stpDump $stpSrc $interval"
		return
	}
	$ns_ at [expr [$ns_ now] + $interval] "$self stpDump $stpSrc $interval"
	puts [$ns_ now]/cwnd=[format "%.4f" [$stpSrc set cwnd_]]/ack=[$stpSrc set highest_ack_]
}

#
# Arrange for tcp source stats to be dumped for $tcpSrc every
# $interval seconds of simulation time
#
TestSuite instproc tcpDump { tcpSrc interval } {
	$self instvar dump_inst_ ns_
	if ![info exists dump_inst_($tcpSrc)] {
		set dump_inst_($tcpSrc) 1
		$ns_ at 0.0 "$self tcpDump $tcpSrc $interval"
		return
	}
	$ns_ at [expr [$ns_ now] + $interval] "$self tcpDump $tcpSrc $interval"
	puts [$ns_ now]/cwnd=[format "%.4f" [$tcpSrc set cwnd_]]/ssthresh=[$tcpSrc set ssthresh_]/ack=[$tcpSrc set ack_]
}

TestSuite instproc tcpDumpAll { tcpSrc interval label } {
	$self instvar dump_inst_ ns_
	if ![info exists dump_inst_($tcpSrc)] {
		set dump_inst_($tcpSrc) 1
		puts $label/window=[$tcpSrc set window_]/packetSize=[$tcpSrc set packetSize_]/bugFix=[$tcpSrc set bugFix_]	
		$ns_ at 0.0 "$self tcpDumpAll $tcpSrc $interval $label"
		return
	}
	$ns_ at [expr [$ns_ now] + $interval] "$self tcpDumpAll $tcpSrc $interval $label"
	puts $label/time=[$ns_ now]/cwnd=[format "%.4f" [$tcpSrc set cwnd_]]/ssthresh=[$tcpSrc set ssthresh_]/ack=[$tcpSrc set ack_]/rtt=[$tcpSrc set rtt_]	
}

TestSuite instproc openTrace { stopTime testName } {
	$self instvar ns_
	exec rm -f out.tr temp.rands
	set traceFile [open out.tr w]
	puts $traceFile "v testName $testName"
	$ns_ at $stopTime \
		"close $traceFile ; $self finish $testName; exit 0"
	return $traceFile
}

TestSuite instproc openTraceSTP { stopTime testName } {
	$self instvar ns_
	exec rm -f out.tr temp.rands
	set traceFile [open out.tr w]
	puts $traceFile "v testName $testName"
	$ns_ at $stopTime \
		"close $traceFile ; $self finish_stp $testName; exit 0"
	return $traceFile
}

TestSuite instproc openTraceInteract { stopTime testName } {
	$self instvar ns_
	exec rm -f out.tr temp.rands
	set traceFile [open out.tr w]
	puts $traceFile "v testName $testName"
	$ns_ at $stopTime \
		"close $traceFile ; $self finish $testName; $self finish_stp $testName; exit 0"
	return $traceFile
}

TestSuite instproc traceQueues { node traceFile } {
	$self instvar ns_
	foreach nbr [$node neighbors] {
		$ns_ trace-queue $node $nbr $traceFile
		[$ns_ link $node $nbr] trace-dynamics $ns_ $traceFile
	}
}

proc usage {} {
	global argv0
	puts stderr "usage: ns $argv0 <tests>"
	puts stderr "Valid tests are: stp sack newreno"
	puts stderr "The output of this test is: "
	puts stderr "fwd_thruput(kb/s) fwd_xfer_efficiency(kb/s) reverse_thruput(kb/s)"
	exit 1
}

proc isProc? {cls prc} {
	if [catch "Object info subclass $cls/$prc" r] {
		global argv0
		puts stderr "$argv0: no such $cls: $prc"
		usage
	}
}

proc get-subclasses {cls pfx} {
	set ret ""
	set l [string length $pfx]

	set c $cls
	while {[llength $c] > 0} {
		set t [lindex $c 0]
		set c [lrange $c 1 end]
		if [string match ${pfx}* $t] {
			lappend ret [string range $t $l end]
		}
		eval lappend c [$t info subclass]
	}
	set ret
}

TestSuite proc runTest {} {
	global argc argv

	switch $argc {
		1 {
			set test $argv
			isProc? Test $test

			set topo ""
		}
		2 {
			set test [lindex $argv 0]
			isProc? Test $test

			set topo [lindex $argv 1]
			isProc? Topology $topo
		}
		default {
			usage
		}
	}
	set t [new Test/$test $topo]
	$t run
}

# Skeleton topology base class
Class SkelTopology

SkelTopology instproc init {} {
    $self next
}

SkelTopology instproc node? n {
    $self instvar node_
    if [info exists node_($n)] {
	set ret $node_($n)
    } else {
	set ret ""
    }
    set ret
}

SkelTopology instproc add-fallback-links {ns nodelist bw delay qtype args} {
   $self instvar node_
    set n1 [lindex $nodelist 0]
    foreach n2 [lrange $nodelist 1 end] {
	if ![info exists node_($n2)] {
	    set node_($n2) [$ns node]
	}
	$ns duplex-link $node_($n1) $node_($n2) $bw $delay $qtype
	foreach opt $args {
	    set cmd [lindex $opt 0]
	    set val [lindex $opt 1]
	    if {[llength $opt] > 2} {
		set x1 [lindex $opt 2]
		set x2 [lindex $opt 3]
	    } else {
		set x1 $n1
		set x2 $n2
	    }
	    $ns $cmd $node_($x1) $node_($x2) $val
	    $ns $cmd $node_($x2) $node_($x1) $val
	}
	set n1 $n2
    }
}


Class NodeTopology/4nodes -superclass SkelTopology

# Create a simple four node topology:
#
#	   s1
#	     \ 
#  8Mb,5ms \   1.5Mb,260ms
#	        r1 --------- k1
#  8Mb,5ms /
#	     /
#	   s2

NodeTopology/4nodes instproc init ns {
    $self next

    $self instvar node_
    set node_(s1) [$ns node]
    set node_(s2) [$ns node] 
    set node_(s3) [$ns node] 
    set node_(s4) [$ns node] 
    set node_(r1) [$ns node]  
    set node_(k1) [$ns node]    
    set node_(d1) [$ns node]
    set node_(d2) [$ns node]
    set node_(d3) [$ns node]
    set node_(d4) [$ns node]
}

#
# Links1 uses 8Mb, 5ms feeders, and a 800Kb 100ms bottleneck.
# Queue-limit on bottleneck is 6 packets.
#
Class Topology/net0 -superclass NodeTopology/4nodes
Topology/net0 instproc init ns {
    $self next $ns
    $self instvar node_

    $ns duplex-link $node_(s1) $node_(r1) 8Mb 5ms DropTail
    $ns duplex-link $node_(s2) $node_(r1) 8Mb 5ms DropTail
    $ns duplex-link $node_(s3) $node_(r1) 8Mb 5ms DropTail
    $ns duplex-link $node_(s4) $node_(r1) 8Mb 5ms DropTail
    $ns duplex-link $node_(k1) $node_(d1) 8Mb 1ms DropTail
    $ns duplex-link $node_(k1) $node_(d2) 8Mb 1ms DropTail
    $ns duplex-link $node_(k1) $node_(d3) 8Mb 1ms DropTail
    $ns duplex-link $node_(k1) $node_(d4) 8Mb 1ms DropTail
    $ns duplex-link $node_(r1) $node_(k1) 1500Kb 260ms DropTail
    $ns queue-limit $node_(r1) $node_(k1) 18
    $ns queue-limit $node_(k1) $node_(r1) 18
    if {[$class info instprocs config] != ""} {
	$self config $ns
    }
}

# Definition of test-suite tests

Class Test/stp -superclass TestSuite
Test/stp instproc init topo {
	$self instvar net_ defNet_ test_
	set net_	$topo
	set defNet_	net0
	set test_	stp
	$self next
}
Test/stp instproc run {} {
	$self instvar ns_ node_ testName_ stp1

	# Set up TCP connection
	set stp1 [$ns_ create-connection STP $node_(s1) STPSink $node_(k1) 0]
	$stp1 set window_ 200
	
	# Set up FTP source
	set ftp1 [$stp1 attach-source FTP]
	$ns_ at 0.0 "$ftp1 start"

	#$self stpDump $stp1 1.0

    # WWW clients/servers.
	# The number of clients and servers were chosen to produce about a
	# 10% average load on the network in the direction in which the throughput
	# is being measured.  The servers and clients are separated into
	# groups because TCP ports are not reused in ns, and too many servers
	# and clients on one TCP agent causes port exhaustion on the agent.
    initCdfs $ns_
    global wwwServer
    set serv 4
    set wwwServer(0) [new WWWServer $node_(s1) $ns_]
    set wwwServer(1) [new WWWServer $node_(s2) $ns_]
    set wwwServer(2) [new WWWServer $node_(s3) $ns_]
    set wwwServer(3) [new WWWServer $node_(s4) $ns_]
    global wwwClient
    set cl_per_serv 9
    for {set i 0} {$i < $cl_per_serv} {incr i} {
        set wwwClient(0,$i) [new WWWClient $node_(d1) $ns_]
        $ns_ at 0.0 "$wwwClient(0,$i) start $wwwServer(0)"
    }
    for {set i 0} {$i < $cl_per_serv} {incr i} {
        set wwwClient(1,$i) [new WWWClient $node_(d2) $ns_]
        $ns_ at 0.0 "$wwwClient(1,$i) start $wwwServer(1)"
    } 
    for {set i 0} {$i < $cl_per_serv} {incr i} {
        set wwwClient(2,$i) [new WWWClient $node_(d3) $ns_]
        $ns_ at 0.0 "$wwwClient(2,$i) start $wwwServer(2)"
    }       
    for {set i 0} {$i < $cl_per_serv} {incr i} {
        set wwwClient(3,$i) [new WWWClient $node_(d4) $ns_]
        $ns_ at 0.0 "$wwwClient(3,$i) start $wwwServer(3)"
    }

	# Trace only the bottleneck link
	#
	# Actually, we now trace all activity at the node around the
	# bottleneck link.  This allows us to track acks, as well
	# packets taking any alternate paths around the bottleneck
	# link.
	#
	$self traceQueues $node_(r1) [$self openTraceSTP 60.0 $testName_]
	$ns_ run
}

Class Test/sack -superclass TestSuite
Test/sack instproc init topo {
	$self instvar net_ defNet_ test_
	set net_	$topo
	set defNet_	net0
	set test_	sack
	$self next
}
Test/sack instproc run {} {
	$self instvar ns_ node_ testName_ tcp1

	set tcp1 [$ns_ create-connection TCP/Sack1 $node_(s1) TCPSink/Sack1/DelAck $node_(k1) 0]
	$tcp1 set window_ 200
	
	set ftp1 [$tcp1 attach-source FTP]
	$ns_ at 0.0 "$ftp1 start"

#	$self tcpDump $tcp1 1.0

    # WWW clients/servers.
    initCdfs $ns_
    global wwwServer
    set serv 4
    set wwwServer(0) [new WWWServer $node_(s1) $ns_]
    set wwwServer(1) [new WWWServer $node_(s2) $ns_]
    set wwwServer(2) [new WWWServer $node_(s3) $ns_]
    set wwwServer(3) [new WWWServer $node_(s4) $ns_]
    global wwwClient
    set cl_per_serv 9
    for {set i 0} {$i < $cl_per_serv} {incr i} {
        set wwwClient(0,$i) [new WWWClient $node_(d1) $ns_]
        $ns_ at 0.0 "$wwwClient(0,$i) start $wwwServer(0)"
    }
    for {set i 0} {$i < $cl_per_serv} {incr i} {
        set wwwClient(1,$i) [new WWWClient $node_(d2) $ns_]
        $ns_ at 0.0 "$wwwClient(1,$i) start $wwwServer(1)"
    } 
    for {set i 0} {$i < $cl_per_serv} {incr i} {
        set wwwClient(2,$i) [new WWWClient $node_(d3) $ns_]
        $ns_ at 0.0 "$wwwClient(2,$i) start $wwwServer(2)"
    }       
    for {set i 0} {$i < $cl_per_serv} {incr i} {
        set wwwClient(3,$i) [new WWWClient $node_(d4) $ns_]
        $ns_ at 0.0 "$wwwClient(3,$i) start $wwwServer(3)"
    }

	# Trace only the bottleneck link
	$self traceQueues $node_(r1) [$self openTrace 60.0 $testName_]
	$ns_ run
}

# This test looks at the interaction between Sack and STP
Class Test/interact -superclass TestSuite
Test/interact instproc init topo {
	$self instvar net_ defNet_ test_
	set net_	$topo
	set defNet_	net0
	set test_	interact	
	$self next
}
Test/interact instproc run {} {
	$self instvar ns_ node_ testName_ tcp1 stp1

	$ns_ queue-limit $node_(r1) $node_(k1) 8   
	$ns_ queue-limit $node_(k1) $node_(r1) 8   

	set tcp1 [$ns_ create-connection TCP/Sack1 $node_(s1) TCPSink/Sack1 $node_(k1) 0]
	$tcp1 set window_ 200
	set stp1 [$ns_ create-connection STP $node_(s1) STPSink $node_(k1) 0]
	$stp1 set window_ 200

	set ftp1 [$tcp1 attach-source FTP]
	set ftp2 [$stp1 attach-source FTP]

	$ns_ at 0.0 "$ftp1 start"
	$ns_ at 0.0 "$ftp2 start"

	$self tcpDump $tcp1 1.0
	$self stpDump $stp1 1.0
    # WWW clients/servers.
    initCdfs $ns_
    global wwwServer
    set serv 4
    set wwwServer(0) [new WWWServer $node_(s1) $ns_]
    set wwwServer(1) [new WWWServer $node_(s2) $ns_]
    set wwwServer(2) [new WWWServer $node_(s3) $ns_]
    set wwwServer(3) [new WWWServer $node_(s4) $ns_]
    global wwwClient
    set cl_per_serv 9
    for {set i 0} {$i < $cl_per_serv} {incr i} {
        set wwwClient(0,$i) [new WWWClient $node_(d1) $ns_]
        $ns_ at 0.0 "$wwwClient(0,$i) start $wwwServer(0)"
    }
    for {set i 0} {$i < $cl_per_serv} {incr i} {
        set wwwClient(1,$i) [new WWWClient $node_(d2) $ns_]
        $ns_ at 0.0 "$wwwClient(1,$i) start $wwwServer(1)"
    } 
    for {set i 0} {$i < $cl_per_serv} {incr i} {
        set wwwClient(2,$i) [new WWWClient $node_(d3) $ns_]
        $ns_ at 0.0 "$wwwClient(2,$i) start $wwwServer(2)"
    }       
    for {set i 0} {$i < $cl_per_serv} {incr i} {
        set wwwClient(3,$i) [new WWWClient $node_(d4) $ns_]
        $ns_ at 0.0 "$wwwClient(3,$i) start $wwwServer(3)"
    }


	# Trace only the bottleneck link
	$self traceQueues $node_(r1) [$self openTraceInteract 60.0 $testName_]
	$ns_ run
}

Class Test/newreno -superclass TestSuite
Test/newreno instproc init topo {
	$self instvar net_ defNet_ test_
	set net_	$topo
	set defNet_	net0
	set test_	newreno
	$self next
}
Test/newreno instproc run {} {
	$self instvar ns_ node_ testName_ tcp1

	set tcp1 [$ns_ create-connection TCP/Newreno $node_(s1) TCPSink/DelAck $node_(k1) 0]
	$tcp1 set window_ 200

	set ftp1 [$tcp1 attach-source FTP]
	$ns_ at 0.0 "$ftp1 start"

#	$self tcpDump $tcp1 1.0
    # WWW clients/servers.
    initCdfs $ns_
    global wwwServer
    set serv 4
    set wwwServer(0) [new WWWServer $node_(s1) $ns_]
    set wwwServer(1) [new WWWServer $node_(s2) $ns_]
    set wwwServer(2) [new WWWServer $node_(s3) $ns_]
    set wwwServer(3) [new WWWServer $node_(s4) $ns_]
    global wwwClient
    set cl_per_serv 9
    for {set i 0} {$i < $cl_per_serv} {incr i} {
        set wwwClient(0,$i) [new WWWClient $node_(d1) $ns_]
        $ns_ at 0.0 "$wwwClient(0,$i) start $wwwServer(0)"
    }
    for {set i 0} {$i < $cl_per_serv} {incr i} {
        set wwwClient(1,$i) [new WWWClient $node_(d2) $ns_]
        $ns_ at 0.0 "$wwwClient(1,$i) start $wwwServer(1)"
    } 
    for {set i 0} {$i < $cl_per_serv} {incr i} {
        set wwwClient(2,$i) [new WWWClient $node_(d3) $ns_]
        $ns_ at 0.0 "$wwwClient(2,$i) start $wwwServer(2)"
    }       
    for {set i 0} {$i < $cl_per_serv} {incr i} {
        set wwwClient(3,$i) [new WWWClient $node_(d4) $ns_]
        $ns_ at 0.0 "$wwwClient(3,$i) start $wwwServer(3)"
    }

	# trace only the bottleneck link
	$self traceQueues $node_(r1) [$self openTrace 60.0 $testName_]
	$ns_ run
}

Agent/STPSink set win_ 200
Agent/STPSink set p_error_ 0

TestSuite runTest

### Local Variables:
### mode: tcl
### tcl-indent-level: 8
### tcl-default-application: ns
### End:
